from bussdcc import ContextProtocol

from ..base import FlaskApp
from ..protocol import WebPlugin
from .database import db


class SqlAlchemyPlugin:
    name = "sqlalchemy"

    def init_app(self, app: FlaskApp, ctx: ContextProtocol) -> None:
        db.init_app(app)


plugin: WebPlugin = SqlAlchemyPlugin()



from __future__ import annotations
from dataclasses import Field, dataclass
from datetime import date, datetime, time
from enum import Enum
from types import MappingProxyType, UnionType
from typing import Any, Literal, Union, get_args, get_origin


def _unwrap_optional(tp: object) -> object:
    origin = get_origin(tp)
    args = get_args(tp)

    if origin in (Union, UnionType):
        non_none = [arg for arg in args if arg is not type(None)]
        if len(non_none) == 1:
            return non_none[0]

    return tp


def _enum_type(tp: object) -> type[Enum] | None:
    if isinstance(tp, type) and issubclass(tp, Enum):
        return tp
    return None


def _field_shape(
    tp: object,
    value: Any | None,
) -> tuple[str, list[str] | None, Any | None]:
    base_tp = _unwrap_optional(tp)
    origin = get_origin(base_tp)
    args = get_args(base_tp)

    enum_tp = _enum_type(base_tp)

    if origin is Literal:
        return "select", [str(v) for v in args], value

    if enum_tp is not None:
        if isinstance(value, Enum):
            value = value.value
        return "select", [str(member.value) for member in enum_tp], value

    if base_tp in (int, float):
        return "number", None, value

    if base_tp is bool:
        return "checkbox", None, value

    if base_tp is date:
        return "date", None, value

    if base_tp is time:
        return "time", None, value

    if base_tp is datetime:
        return "datetime-local", None, value

    return "text", None, value


@dataclass(slots=True, frozen=True)
class FieldMetadata:
    label: str
    group: str = "General"
    required: bool = False
    help: str | None = None
    min: int | float | None = None
    max: int | float | None = None
    step: int | float | None = None

    @staticmethod
    def from_field(f: Field[object]) -> "FieldMetadata":
        meta: MappingProxyType[Any, Any] = f.metadata

        return FieldMetadata(
            label=meta.get("label", f.name),
            group=meta.get("group", "General"),
            required=meta.get("required", False),
            help=meta.get("help"),
            min=meta.get("min"),
            max=meta.get("max"),
            step=meta.get("step"),
        )


@dataclass(slots=True, frozen=True)
class TreeField:
    name: str
    type: object
    meta: FieldMetadata
    value: Any | None = None
    input_type: str | None = None
    options: list[str] | None = None

    @staticmethod
    def create(
        name: str,
        type: object,
        value: Any | None = None,
        *,
        label: str | None = None,
        required: bool = False,
        help: str | None = None,
        min: int | float | None = None,
        max: int | float | None = None,
        step: int | float | None = None,
    ) -> "TreeField":
        meta = FieldMetadata(
            label=label or name,
            required=required,
            help=help,
            min=min,
            max=max,
            step=step,
        )

        input_type, options, value = _field_shape(type, value)

        return TreeField(
            name=name,
            type=type,
            meta=meta,
            value=value,
            input_type=input_type,
            options=options,
        )

    @staticmethod
    def from_field(f: Field[object], value: Any | None = None) -> "TreeField":
        meta = FieldMetadata.from_field(f)
        input_type, options, value = _field_shape(f.type, value)

        return TreeField(
            name=f.name,
            type=f.type,
            meta=meta,
            value=value,
            input_type=input_type,
            options=options,
        )


@dataclass(slots=True, frozen=True)
class TreeMappingEntry:
    name: str
    key: TreeField
    value: TreeField | TreeNode


@dataclass(slots=True, frozen=True)
class TreeMapping:
    name: str
    entries: list[TreeMappingEntry]
    key_schema: TreeField
    value_schema: TreeField | TreeNode


@dataclass(slots=True, frozen=True)
class TreeListEntry:
    name: str
    value: TreeField | TreeNode


@dataclass(slots=True, frozen=True)
class TreeList:
    name: str
    entries: list[TreeListEntry]
    item_schema: TreeField | TreeNode


@dataclass(slots=True, frozen=True)
class TreeNode:
    name: str | None
    fields: list[TreeField]
    children: list["TreeNode"]
    mappings: list[TreeMapping]
    lists: list[TreeList]



from typing import Any, get_args, get_origin, Literal, Union
from datetime import date, time, datetime
from enum import Enum
from types import UnionType


def unwrap_optional(tp: Any) -> Any:
    origin = get_origin(tp)
    args = get_args(tp)

    if origin in (Union, UnionType):
        non_none = [arg for arg in args if arg is not type(None)]
        if len(non_none) == 1:
            return non_none[0]

    return tp


def coerce_form_value(tp: Any, raw: Any) -> Any:
    tp = unwrap_optional(tp)

    if tp is bool:
        return raw in ("on", "true", "1", "yes", True)

    if raw == "":
        if tp is str:
            return ""
        return None

    if raw is None:
        return None

    origin = get_origin(tp)

    if origin is Literal:
        return raw

    if isinstance(tp, type) and issubclass(tp, Enum):
        return raw

    if tp is int:
        return int(raw)

    if tp is float:
        return float(raw)

    if tp in (date, time, datetime):
        return raw

    return raw



from flask import Blueprint

from bussdcc import ContextProtocol

from ..base import FlaskApp
from ..protocol import WebPlugin
from .types import TreeNode


class FormtreePlugin:
    name = "formtree"

    def init_app(self, app: FlaskApp, ctx: ContextProtocol) -> None:
        app.jinja_env.tests["tree_node"] = lambda x: isinstance(x, TreeNode)

        bp = Blueprint(
            "bussdcc_framework_formtree",
            __name__,
            template_folder="templates",
            static_folder="static",
            static_url_path="/_framework/formtree/static",
        )
        app.register_blueprint(bp)


plugin: WebPlugin = FormtreePlugin()



from flask import Blueprint

from bussdcc import ContextProtocol

from ..base import FlaskApp
from ..protocol import WebPlugin


class SocketIOPlugin:
    name = "socketio"

    def init_app(self, app: FlaskApp, ctx: ContextProtocol) -> None:
        bp = Blueprint(
            "bussdcc_framework_socketio",
            __name__,
            template_folder="templates",
            static_folder="static",
        )

        app.config.setdefault("SOCKETIO_RECONNECT_DELAY", 1000)
        app.config.setdefault("SOCKETIO_RECONNECT_DELAY_MAX", 5000)

        app.register_blueprint(bp)


plugin: WebPlugin = SocketIOPlugin()



import traceback
from typing import Optional
from datetime import datetime

from bussdcc.runtime.replay import ReplayRuntime as Base
from bussdcc.clock import ClockProtocol
from bussdcc.io import EventSinkProtocol
from bussdcc.state import StateStoreProtocol
from bussdcc.event import EventBusProtocol
from bussdcc import Event, Message, Severity

from .. import message, __version__ as version


class ReplayRuntime(Base):
    def __init__(
        self,
        *,
        clock: Optional[ClockProtocol] = None,
        events: EventBusProtocol | None = None,
        state: StateStoreProtocol | None = None,
        speed: float = 1.0,
        start_at: datetime | None = None,
    ):
        super().__init__(
            clock=clock, events=events, state=state, speed=speed, start_at=start_at
        )
        self._sinks: list[EventSinkProtocol] = []

    def boot(self) -> None:
        for sink in self._sinks:
            sink.start(self.ctx)

        super().boot()

        self.ctx.emit(message.FrameworkBooted(version=version))

    def _dispatch(self, evt: Event[Message]) -> None:
        for sink in self._sinks:
            try:
                sink.handle(evt)
            except Exception as e:
                if evt.payload.severity >= Severity.ERROR:
                    continue

                self.ctx.emit(
                    message.SinkHandlerError(
                        sink=type(sink).__name__,
                        error=repr(e),
                        evt=evt,
                        traceback=traceback.format_exc(),
                    )
                )

        super()._dispatch(evt)

    def shutdown(self, reason: Optional[str] = None) -> None:
        self.ctx.emit(message.FrameworkShuttingDown(version=version, reason=reason))

        try:
            super().shutdown(reason)
        finally:
            for sink in self._sinks:
                try:
                    sink.stop()
                except Exception:
                    pass

    def add_sink(self, sink: EventSinkProtocol) -> None:
        if self._booted:
            raise RuntimeError("Cannot add sinks after boot")
        self._sinks.append(sink)



from flask_bootstrap import Bootstrap5  # type: ignore[import-untyped]

from bussdcc import ContextProtocol

from ..base import FlaskApp
from ..protocol import WebPlugin


class BootstrapPlugin:
    name = "bootstrap"

    def init_app(self, app: FlaskApp, ctx: ContextProtocol) -> None:
        Bootstrap5(app)


plugin: WebPlugin = BootstrapPlugin()



from flask import Blueprint

from bussdcc import ContextProtocol

from ..base import FlaskApp
from ..protocol import WebPlugin


class BootstrapLayoutPlugin:
    name = "bootstrap-layout"

    def init_app(self, app: FlaskApp, ctx: ContextProtocol) -> None:
        bp = Blueprint(
            "bussdcc_framework_bootstrap_layout",
            __name__,
            template_folder="templates",
        )
        app.register_blueprint(bp)


plugin: WebPlugin = BootstrapLayoutPlugin()
